KeyRunner = LibStub("AceAddon-3.0"):NewAddon("KeyRunner", "AceEvent-3.0")
LSM = LibStub:GetLibrary("LibSharedMedia-3.0");
LSM:Register("font", "Gotham Ultra", "Interface\\Addons\\KeyRunner\\Media\\gotham-ultra.ttf")


function KeyRunner:OnInitialize()
    -- Called when the addon is loaded
    -- Register Frames
  KeyRunnerFrame = CreateFrame("Frame", "KeyRunnerFrame", UIParent);
  CompareTooltipFrame = CreateFrame("Frame","CompareTooltipFrame",KeyRunnerFrame)
  EmptyHideFrame = CreateFrame("Frame");
  EmptyHideFrame:Hide();
  bgTexture = KeyRunnerFrame:CreateTexture(nil,"BACKGROUND");
 
end

function KeyRunner:OnEnable()
    -- Called when the addon is enabled
    -- Register Events
    
  self:RegisterEvent("CHALLENGE_MODE_COMPLETED");
  self:RegisterEvent("CHALLENGE_MODE_RESET");
  self:RegisterEvent("CHALLENGE_MODE_START")
  self:RegisterEvent("PLAYER_ENTERING_WORLD")
  self:RegisterEvent("SCENARIO_CRITERIA_UPDATE");
  self:RegisterEvent("WORLD_STATE_TIMER_START");
  self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
  self:RegisterEvent("ENCOUNTER_START")
  self:RegisterEvent("ENCOUNTER_END")
  self:RegisterEvent("PLAYER_LOGOUT")
    
  hooksecurefunc("Scenario_ChallengeMode_UpdateTime", updateTime)
	
  if (KeyRunnerConfigDB == nil) then 
    KeyRunnerConfigDB = {}
  end
	  
  if (KeyRunnerKeysDB == nil) then 
    KeyRunnerKeysDB = {}
  end
  
  local currentVersionKeysDB = 1.1
  local currentVersionClient = 1.2 
  if getConfig("versionKeysDB") < currentVersionKeysDB then
    updateKeysDB(currentVersionKeysDB)
    setConfig("versionKeysDB" , currentVersionKeysDB )
  end
  setConfig("versionClient",currentVersionClient)  
  
  bgTexture:SetAllPoints(KeyRunnerFrame);
  bgTexture:SetColorTexture(0.5, 0.5, 0.5, 0);
  toggleFrameMovement()
  positionFrame()
	
  -- init globals
  currentKey = getConfig("currentKey")
  loadCompareKey()
  demoKey = nil
  demoCompareKey = nil  
  optionsShown = false
  keyShown = false
  demoShown = false
  countString1 = ""
  countString2 = ""
  countString3 = ""
  
  createFontStringObjects() 
  setFontStringsAppearance()
  createBlizOptions()
	
  CompareTooltipFrame:SetScript("OnEnter", setCompareTooltip)
  CompareTooltipFrame:SetScript("OnLeave", hideTooltip)
	
  SlashCmdList['MYADDON_SLASHCMD'] = function(msg)
    if not optionsShown then 
      createBaseContainer() 
	end
  end
  SLASH_MYADDON_SLASHCMD1 = '/KeyRunner'
		 
		 
end



function KeyRunner:OnDisable() -- Called when the addon is disabled

  setConfig("currentKey",currentKey)
  
end


function frameStopMoving()
  KeyRunnerFrame:StopMovingOrSizing()

  local _, _, relativePoint, xOffset, yOffset = KeyRunnerFrame:GetPoint() -- relative point is to closest point

  local framePosition = {}
  framePosition.xOffset = xOffset
  framePosition.yOffset = yOffset
  framePosition.relativePoint = relativePoint

  setConfig("framePosition",framePosition)

end

function unlockFrame()

  KeyRunnerFrame:SetMovable(true)
  KeyRunnerFrame:EnableMouse(true)
  KeyRunnerFrame:RegisterForDrag("LeftButton")
  KeyRunnerFrame:SetScript("OnDragStart", KeyRunnerFrame.StartMoving)
  KeyRunnerFrame:SetScript("OnDragStop", frameStopMoving)
	
end


function toggleFrameMovement()

  if getConfig("frameUnlocked") then 
    unlockFrame()
  else
    KeyRunnerFrame:SetMovable(false)
    KeyRunnerFrame:EnableMouse(false)
  end
  
end

function positionFrame()

  KeyRunnerFrame:ClearAllPoints() 
	
  local framePosition = getConfig("framePosition")
  KeyRunnerFrame:SetPoint(framePosition.relativePoint,framePosition.xOffset, framePosition.yOffset) 
  KeyRunnerFrame:SetWidth(200)
  KeyRunnerFrame:SetHeight(200)
  
end


-- Events

function KeyRunner:PLAYER_LOGOUT() -- before Disconnect/logout/reload ui and right before savedVariables gets saved

  setConfig("currentKey",currentKey)
  
end

function KeyRunner:CHALLENGE_MODE_COMPLETED()
  
  if currentKey then 
    currentKey.completed = true
    if getConfig("compareKeySave").mode == 1 then autoSaveKey() end
  end
 
end

function KeyRunner:CHALLENGE_MODE_RESET() -- fired right before new key started
  
  if currentKey then 
    currentKey = nil 
    KeyRunnerFrame:Hide() 
    keyShown = false
  end

end

function KeyRunner:CHALLENGE_MODE_START() -- fired twice sometimes at start because not loaded yet
  -- fired every time rezone into key and also on relog but not on reload ui
   
  if  currentKey then --  rezoning/relog
  
  else 
    initKey()
    loadCompareKey()
    updateSteps()
 
  end
 
  resolveFrames()
 
end

function KeyRunner:PLAYER_ENTERING_WORLD() -- fired on reload UI and relog and rezone into key

  if isKeyActive() then  
    currentKeyID,currentElapsedTime,count,bossesKilled = getActiveKeyData()
    if currentKey == nil or currentKey.completed or currentKey.keyID ~= currentKeyID  or currentKey.elapsedTime>currentElapsedTime or currentKey.currentCount>count or getBossesKilledCount(currentKey.bossInfo) > bossesKilled  then -- addon was enabled after key was started
      currentKey = nil 
	  initKey()
	  loadCompareKey()
    end
  
    recoverSteps()
	updateSteps()
 
  end
  resolveFrames()
  
end


function KeyRunner:SCENARIO_CRITERIA_UPDATE() -- fires a lot of times after the last step for some reason 

  updateSteps()
 
end


function KeyRunner:WORLD_STATE_TIMER_START() -- at start of key and after every death

  updateSteps() 
   
end



function KeyRunner:ZONE_CHANGED_NEW_AREA()

  resolveFrames()
 
end




function KeyRunner:ENCOUNTER_START() 

  if currentKey and isKeyActive() then
    currentKey["lastBossStartTime"] = math.floor(GetTime()) -- time since your own computer booted up. 
  end
 
end

function KeyRunner:ENCOUNTER_END()

  if currentKey and isKeyActive() then
    currentKey["lastBossEndTime"] = math.floor(GetTime()) -- time since your own computer booted up. 
  end
 
end





function createDemoKey()

  demoKey = {};
 
  demoKey["mapName"], demoKey["keyID"], demoKey["timer"] = "Freehold", 4561, 2400
  demoKey["elapsedTime"] = 1666;
  demoKey["zoneID"] = 246;
  demoKey["deaths"] = 5;
  demoKey["numSteps"] = 5; 
  demoKey["bossCount"] = 4
  demoKey["level"] = 21;
  demoKey["completed"] = false;
  demoKey["affixIDs"] = {2,7,10};
  demoKey["bossInfo"] = {   {["name"] = "The Sand Queen defeated",["killTime"] =300, ["duration"] = 120, ["completed"] = true},
                            {["name"] = "Jes Howlis defeated",["killTime"] = 1205, ["duration"] = 145, ["completed"] = true},
						    {["name"] = "Knight Captain Valyri defeated",["killTime"] = 0, ["duration"] = 0, ["completed"]= false},
							{["name"] = "Overseer Korgus defeated",["killTime"] = 1666, ["duration"] = 120, ["completed"]= true},
						};
  demoKey["finalCount"] = 214;
  demoKey["currentCount"] = 80;
  demoKey["countInfo"] = {{0,0},{4,20},{9,35},{20,60},{80,120}};
  demoKey["missingCountInfo"] = {
    ["missingCount"] = {},
    ["missingTime"] = {},
  }
  demoKey["dataMissing"] = false
  demoKey["lastCountUpdate"] = 120
  
  

end


function createDemoCompareKey()

  demoCompareKey = {};
 
  demoCompareKey["mapName"], demoCompareKey["keyID"], demoCompareKey["timer"] = "Freehold", 4561, 2400
  demoCompareKey["elapsedTime"] = 2200;
  demoCompareKey["zoneID"] = 246;
  demoCompareKey["deaths"] = 7;
  demoCompareKey["numSteps"] = 5; 
  demoCompareKey["bossCount"] = 4
  demoCompareKey["level"] = 20;
  demoCompareKey["completed"] = true;
  demoCompareKey["affixIDs"] = {2,7,10};
  demoCompareKey["bossInfo"] = {
                                {["name"] = "The Sand Queen defeated",["killTime"] = 160, ["duration"] = 125, ["completed"] = true},
                                {["name"] = "Jes Howlis defeated",["killTime"] = 1206, ["duration"] = 160, ["completed"] = true},
						        {["name"] = "Knight Captain Valyri defeated",["killTime"] = 1500, ["duration"] = 175, ["completed"] = true},
								{["name"] = "Overseer Korgus defeated",["killTime"] = 2100, ["duration"] = 30, ["completed"]= true},
						       };
  demoCompareKey["finalCount"] = 214;
  demoCompareKey["currentCount"] = 214;
  demoCompareKey["countInfo"] = {{0,0},{8,35},{10,35},{40,60},{120,160},{214,2200}};
  demoCompareKey["missingCountInfo"] = {
    ["missingCount"] = {},
    ["missingTime"] = {},
  }
  demoCompareKey["dataMissing"] = false
  demoCompareKey["lastCountUpdate"] = 2200
 
  
end


function initKey() 
  
  
  
  local _, _, _, _, _, _, _, zoneID = GetInstanceInfo(); 
  local _, _, numSteps = C_Scenario.GetStepInfo(); 
  local keyLevel, affixIDs, multiplier = C_ChallengeMode.GetActiveKeystoneInfo();
  local keyID = C_ChallengeMode.GetActiveChallengeMapID();
 
  
  currentKey = {};
 
  currentKey["mapName"], currentKey["keyID"], currentKey["timer"] = C_ChallengeMode.GetMapUIInfo(keyID); 
  currentKey["elapsedTime"] = 0;
  currentKey["zoneID"] = zoneID;
  currentKey["deaths"] = 0;
  currentKey["numSteps"] = 0 
  currentKey["bossCount"] = 0
  currentKey["level"] = keyLevel;
  currentKey["score"] = 0
  currentKey["isSetCompare"] = false
  currentKey["completed"] = false;
  currentKey["affixIDs"] = affixIDs;
  currentKey["bossInfo"] = {}; -- cant init table here because certain dungeons like shrine dont get step info till timer starts
  currentKey["lastBossStartTime"]  = 0  
  currentKey["lastBossEndTime"] = 0
  currentKey["finalCount"] = 0;
  currentKey["currentCount"] = 0;
  currentKey["dataMissing"] = false
  currentKey["lastCountUpdate"] = 0
  currentKey["date"] = date("%m/%d/%y")
  currentKey["time"] = date("%H:%M")
  currentKey["players"] = getPlayers()
  currentKey["missingCountInfo"] = {
    ["missingCount"] = {},
    ["missingTime"] = {},
  }
  currentKey["countInfo"] = {{0,0}};
  
 
	
end


function getPlayer(unit)
  
  local player = {}
  player.name = GetUnitName(unit, false);
  _ , player.class = UnitClass(unit);
  player.role = UnitGroupRolesAssigned(unit);
  player.deaths = 0
  
  return player

end

function getPlayers()

  local players = {}
  local groupCount = GetNumGroupMembers();
  local prefix = "party"
  if IsInRaid() then prefix = "raid" end
   
  for i = 1, groupCount - 1 do
    if UnitExists(prefix..i) then table.insert(players, getPlayer(prefix..i)) end
  end
  table.insert(players, getPlayer("player"));
 
  return players

end

function replaceWorstKey(mapName)

  local index = 0
  local minScore = 0

  if currentKey ~=nil then 
    local keyList = getKeysConfig(mapName)
    if next(keyList) then
      for i, key in ipairs(keyList) do
	    key.score = getKeyScore(key)
        local score = key.score
		if(index==0) then 
	      index = i
		  minScore = score
		else 
	      if(score<=minScore) then 
		    index = i
			minScore = score
		  end
		end
      end
    end
  end
  
  if index > 0 and currentKey ~=nil then 
    currentKey.score = getKeyScore(currentKey)
    if currentKey.score >= minScore then 
      local keyList = getKeysConfig(mapName)
      keyList[index] = currentKey
      setKeysConfig(mapName , keyList)
    end
  end	
	
end

function getBestKey(mapName)

  local index = 0
  local maxScore = 0
  local returnKey = nil
    
  local keyList = getKeysConfig(mapName)
  if next(keyList) then
    for i, key in ipairs(keyList) do
	  if key.dataMissing ~= nil and key.dataMissing then 
		   
	  else
	    key.score = getKeyScore(key)
        local score = key.score
		if (index == 0) then 
	      index = i 
		  maxScore = score
		else
		  if(score>=maxScore) then 
		    index = i
			maxScore = score
		  end
	    end
      end
    end
  end
	
  if index > 0 then 
    local keyList = getKeysConfig(currentKey.mapName)
	if keyList[index] then returnKey = keyList[index] end
  end
  return returnKey
  
end

function isKeysMax(mapName) 

  local maxKeys = getConfig("compareKeySave").max
  local keysCount = 0
   
  local keyList = getKeysConfig(mapName)
  if next(keyList) then
    for i, value in ipairs(keyList) do
      keysCount = keysCount + 1
    end
  end
    
  if keysCount < maxKeys  then 
    return false
  else 
    return true
  end
   
end
  
function getKeyScore(key)

  local returnScore = 0
  local t = findTime(key.finalCount,key.countInfo)
  if key.completed and key.finalCount>0 and t and key.bossCount>0 and getBossesKilledCount(key.bossInfo)==key.bossCount then 
    local percent = (key.timer - key.elapsedTime)/key.timer*100
    percent = tonumber(string.format("%.2f", percent))
    if percent < 0 then 
	  percent = percent * 5
	end
    returnScore = key.level * 25 + percent
  else
    local countScore = 0
    if key.finalCount >0 then 
      countScore = tonumber(string.format("%.2f",key.currentCount/key.finalCount))
    end
    returnScore = key.level + countScore
  end
  return returnScore
end

function saveKey()

  if currentKey then 
    if not isKeysMax(currentKey.mapName) then
      local keyList = getKeysConfig(currentKey.mapName)
      table.insert(keyList, currentKey)
      setKeysConfig(currentKey.mapName , keyList)
    end
  end
 
end



function autoSaveKey() 

  if currentKey then 
    if not isKeysMax(currentKey.mapName) then
      saveKey()
      currentKey = nil
    else 
	  replaceWorstKey(currentKey.mapName)
	  currentKey = nil  
	end
  end
  
end




function setCompareKey(mapName,index)
 
  local keyList = getKeysConfig(mapName)
   
  for i, key in ipairs(keyList) do
    if i == index then 
	  key.isSetCompare = true
	else 
	  key.isSetCompare = false
	end
  end
  setKeysConfig(mapName,keyList)
  
end

function getCompareKeyIndex(mapName)

  local returnIndex = 0
  local keyList = getKeysConfig(mapName)
  for i, key in ipairs(keyList) do
    if  key.isSetCompare then
	  returnIndex = i
	  return returnIndex
	end
  end
  return returnIndex
  
end

function getCompareKey(mapName)

  local localCompareKey = nil
  local keyList = getKeysConfig(mapName)

  for i, key in ipairs(keyList) do
    if  key.isSetCompare then
	  localCompareKey = key
	  return localCompareKey
	end
  end
  return localCompareKey
  
end


function loadCompareKey()  -- sets global compareKey 

  local localCompareKey = nil

  if currentKey ~= nil then 
    local compareKeyLoad = getConfig("compareKeyLoad")
    local keyList = getKeysConfig(currentKey.mapName) 

    if not compareKeyLoad.disabled  and next(keyList)then
      if compareKeyLoad.mode == 1 then -- auto
        localCompareKey = getBestKey(currentKey.mapName)
      else -- manual
        localCompareKey = getCompareKey(currentKey.mapName)
	  end
    end
  end	
  compareKey = localCompareKey
  refreshFrame()

end



function deleteKey(mapName,index)

  local keyList = getKeysConfig(mapName)

  if keyList[index] ~= nil then
    table.remove(keyList,index)
  end

  setKeysConfig(mapName,keyList)

end


function createFontStringObjects () 

  if not fontStringTable then 
    fontStringTable = {}
    fontStringTable.timerFontString = KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
    fontStringTable.affixFontString = KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
    fontStringTable.increaseTimersFontString = KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
    fontStringTable.deathsFontString = KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
    fontStringTable.bossesFontString= KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
    fontStringTable.countFontString = KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
    fontStringTable.checkpointsFontString = KeyRunnerFrame:CreateFontString(nil, "OVERLAY");
  end
  
end


function setFontStringsAppearance()

  if not fontStringTable then 
    createFontStringObjects () 
  end
    
  local fonts = getConfig("fonts")

  local alignTop = "TOPRIGHT"
  local alignBot = "BOTTOMRIGHT"

  local flag = {[1] = "OUTLINE" , [2] = "MONOCHROME" , [3] = "THICKOUTLINE" , [4] = "NONE"}
  local alignText =  {[1] = "LEFT" , [2] = "CENTER" , [3] = "RIGHT" }

  if fonts.align == 1 then 
    alignTop =  "TOPLEFT"
    alignBot =  "BOTTOMLEFT"
  elseif fonts.align == 2 then
    alignTop =  "TOP"
    alignBot =  "BOTTOM"
  elseif fonts.align == 3 then 
    alignTop =  "TOPRIGHT"
    alignBot =  "BOTTOMRIGHT"
  end

  local timerFontSize = fonts.fonts1.size
  local timerFont = LSM:Fetch("font", fonts.fonts1.fontKey)
  local timerFontFlags = flag[fonts.fonts1.flag]
  local timerFontColor = { fonts.fonts1.r, fonts.fonts1.g, fonts.fonts1.b, fonts.fonts1.a }
  
  local affixFontSize =  fonts.fonts2.size
  local affixFont = LSM:Fetch("font", fonts.fonts2.fontKey)
  local affixFontFlags = flag[fonts.fonts2.flag]
  local affixFontColor = { fonts.fonts2.r, fonts.fonts2.g, fonts.fonts2.b, fonts.fonts2.a }
  
  local increaseTimerFontSize = fonts.fonts3.size
  local increaseTimerFont = LSM:Fetch("font", fonts.fonts3.fontKey)
  local increaseTimerFontFlags = flag[fonts.fonts3.flag]
  local increaseTimerFontColor = { fonts.fonts3.r, fonts.fonts3.g, fonts.fonts3.b, fonts.fonts3.a }
  
  local deathsFontSize = fonts.fonts4.size
  local deathsFont =  LSM:Fetch("font", fonts.fonts4.fontKey)
  local deathsFontFlags = flag[fonts.fonts4.flag]
  local deathsFontColor = { fonts.fonts4.r, fonts.fonts4.g, fonts.fonts4.b, fonts.fonts4.a }
  
  local bossFontSize = fonts.fonts5.size
  local bossFont = LSM:Fetch("font", fonts.fonts5.fontKey)
  local bossFontFlags = flag[fonts.fonts5.flag]
  local bossFontColor = { fonts.fonts5.r, fonts.fonts5.g, fonts.fonts5.b, fonts.fonts5.a }
  
  local countFontSize = fonts.fonts6.size
  local countFont =  LSM:Fetch("font", fonts.fonts6.fontKey)
  local countFontFlags = flag[fonts.fonts6.flag]
  local countFontColor = { fonts.fonts6.r, fonts.fonts6.g, fonts.fonts6.b, fonts.fonts6.a }
  
  
  local checkpointFontSize = fonts.fonts7.size
  local checkpointFont =  LSM:Fetch("font", fonts.fonts7.fontKey)
  local checkpointFontFlags = flag[fonts.fonts7.flag]
  local checkpointFontColor = { fonts.fonts7.r, fonts.fonts7.g, fonts.fonts7.b, fonts.fonts7.a }
  
  fontStringTable.timerFontString:SetTextColor(timerFontColor[1],timerFontColor[2], timerFontColor[3], timerFontColor[4]);
  fontStringTable.timerFontString:SetFont(timerFont, timerFontSize, timerFontFlags)
  fontStringTable.timerFontString:ClearAllPoints()
  fontStringTable.timerFontString:SetPoint(alignTop,0,0)

  fontStringTable.affixFontString:SetTextColor(affixFontColor[1],affixFontColor[2], affixFontColor[3], affixFontColor[4]);
  fontStringTable.affixFontString:SetFont(affixFont, affixFontSize, affixFontFlags)
  fontStringTable.affixFontString:ClearAllPoints()
  fontStringTable.affixFontString:SetPoint(alignTop,fontStringTable.timerFontString,alignBot,0,-4)

  CompareTooltipFrame:ClearAllPoints()
  CompareTooltipFrame:SetAllPoints(fontStringTable.affixFontString)

  fontStringTable.increaseTimersFontString:SetTextColor(increaseTimerFontColor[1],increaseTimerFontColor[2], increaseTimerFontColor[3], increaseTimerFontColor[4]);
  fontStringTable.increaseTimersFontString:SetFont(increaseTimerFont, increaseTimerFontSize, increaseTimerFontFlags)
  fontStringTable.increaseTimersFontString:ClearAllPoints()
  fontStringTable.increaseTimersFontString:SetPoint(alignTop,fontStringTable.affixFontString,alignBot,0,-4)

  fontStringTable.deathsFontString:SetTextColor(deathsFontColor[1],deathsFontColor[2], deathsFontColor[3], deathsFontColor[4]);
  fontStringTable.deathsFontString:SetFont(deathsFont, deathsFontSize, deathsFontFlags)
  fontStringTable.deathsFontString:ClearAllPoints()
  fontStringTable.deathsFontString:SetPoint(alignTop,fontStringTable.increaseTimersFontString,alignBot,0,-4)

  fontStringTable.bossesFontString:SetTextColor(bossFontColor[1],bossFontColor[2], bossFontColor[3], bossFontColor[4]);
  fontStringTable.bossesFontString:SetFont(bossFont, bossFontSize, bossFontFlags)
  fontStringTable.bossesFontString:SetSpacing(1)
  fontStringTable.bossesFontString:SetJustifyH(alignText[fonts.align])
  fontStringTable.bossesFontString:ClearAllPoints()
  fontStringTable.bossesFontString:SetPoint(alignTop,fontStringTable.deathsFontString,alignBot,0,-10)  
   
  fontStringTable.countFontString:SetTextColor(countFontColor[1],countFontColor[2], countFontColor[3], countFontColor[4]);
  fontStringTable.countFontString:SetFont(countFont, countFontSize, countFontFlags)
  fontStringTable.countFontString:ClearAllPoints()
  fontStringTable.countFontString:SetPoint(alignTop,fontStringTable.bossesFontString,alignBot,0,-10)

  fontStringTable.checkpointsFontString:SetTextColor(checkpointFontColor[1],checkpointFontColor[2], checkpointFontColor[3], checkpointFontColor[4]);
  fontStringTable.checkpointsFontString:SetFont(checkpointFont, checkpointFontSize, checkpointFontFlags)
  fontStringTable.checkpointsFontString:SetJustifyH(alignText[fonts.align])
  fontStringTable.checkpointsFontString:SetSpacing(1)
  fontStringTable.checkpointsFontString:ClearAllPoints()
  fontStringTable.checkpointsFontString:SetPoint(alignTop,fontStringTable.countFontString,alignBot,0,-1)
  

end

function setFontStringTextsNil()

  if fontStringTable and next(fontStringTable) then  
    fontStringTable.timerFontString:SetText("")
    fontStringTable.affixFontString:SetText("")
    fontStringTable.increaseTimersFontString:SetText("")
    fontStringTable.deathsFontString:SetText("")
    fontStringTable.bossesFontString:SetText("")
    fontStringTable.countFontString:SetText("")
    fontStringTable.checkpointsFontString:SetText("")
  end

end



function updateTimeTexts()

  if currentKey then 
    local timerString = getTimerString(currentKey.elapsedTime,currentKey.timer)
    local affixString = getAffixString(currentKey.level,currentKey.affixIDs)
    local bonusTimerString = getBonusTimerString(currentKey.elapsedTime, currentKey.timer)
    local deathString = getDeathString(currentKey.deaths)
	
	if compareKey then 
	  affixString = affixString .. " " .. getCompareAffixString(compareKey.level)
    end
		 
	if compareKey and getConfig("compareKeyLoad").count then 
	  countString2 = getCompareFindCountString(currentKey.elapsedTime,compareKey.countInfo,compareKey.missingCountInfo,compareKey.finalCount)
	else 
	  countString2 = ""
	end
		 
		
	if fontStringTable and next(fontStringTable) then 
	  fontStringTable.timerFontString:SetText("")
      fontStringTable.affixFontString:SetText("")
      fontStringTable.increaseTimersFontString:SetText("")
      fontStringTable.deathsFontString:SetText("")
	  fontStringTable.countFontString:SetText("")
	 
	  fontStringTable.timerFontString:SetText(timerString)
      fontStringTable.affixFontString:SetText(affixString)
      fontStringTable.increaseTimersFontString:SetText(bonusTimerString)
      fontStringTable.deathsFontString:SetText(deathString)
	  fontStringTable.countFontString:SetText(countString1..countString2..countString3)
      
    end
  end 

end

function updateStepsTexts()
       
  if currentKey then 
    local fullBossString = ""
	countString1 = getCountString(currentKey.currentCount,currentKey.finalCount)
	if compareKey and compareKey.bossInfo and next(compareKey.bossInfo) then
      fullBossString = getFullBossString(currentKey.bossInfo,compareKey.bossInfo,getBossTimerColor(), getConfig("showBossDurations"))
	else 
	  fullBossString = getFullBossString(currentKey.bossInfo,nil,getBossTimerColor(), getConfig("showBossDurations"))
	end
	 
	local checkpointsString = "" 
	 
	if compareKey then 
      checkpointsString = getFullCheckpointString(currentKey,compareKey,getCheckpointTimerColor())
      countString3 = getCompareFindTimeString(currentKey.currentCount,currentKey.missingCountInfo,compareKey.countInfo,compareKey.missingCountInfo,currentKey.lastCountUpdate)
    else
	  checkpointsString = getFullCheckpointString(currentKey,nil,getCheckpointTimerColor())
	  countString3 = ""
    end
	 
	if fontStringTable and next(fontStringTable) then 
	  fontStringTable.bossesFontString:SetText("")
      fontStringTable.countFontString:SetText("")
	  fontStringTable.checkpointsFontString:SetText("")
	  
	  fontStringTable.bossesFontString:SetText(fullBossString)
      fontStringTable.countFontString:SetText(countString1..countString2..countString3)
	  fontStringTable.checkpointsFontString:SetText(checkpointsString)
      
    end
	 
  end

end

function setFontStringsTexts(key1,key2) --key1 is key, key2 is compare

  if key1 then 

    local timerString = getTimerString(key1.elapsedTime,key1.timer)
    local affixString = getAffixString(key1.level,key1.affixIDs)
    local bonusTimerString = getBonusTimerString(key1.elapsedTime, key1.timer)
    local deathString = getDeathString(key1.deaths)
    local fullBossString = ""
    local checkpointsString = "" 

    countString1 = getCountString(key1.currentCount,key1.finalCount)

    if key2 and key2.bossInfo and next(key2.bossInfo) then
      fullBossString = getFullBossString(key1.bossInfo,key2.bossInfo,getBossTimerColor(), getConfig("showBossDurations"))
	else 
	  fullBossString = getFullBossString(key1.bossInfo,nil,getBossTimerColor(), getConfig("showBossDurations"))
	end

    if key2 then 
      affixString = affixString .. " " .. getCompareAffixString(key2.level)
      checkpointsString = getFullCheckpointString(key1,key2,getCheckpointTimerColor()) 
      countString3 = getCompareFindTimeString(key1.currentCount,key1.missingCountInfo,key2.countInfo,key2.missingCountInfo,key1.lastCountUpdate)
    else 
	  checkpointsString = getFullCheckpointString(key1,key2,getCheckpointTimerColor())
	  countString3 = ""
	end
	
	if key2 and getConfig("compareKeyLoad").count then 
	  countString2 = getCompareFindCountString(key1.elapsedTime,key2.countInfo,key2.missingCountInfo,key2.finalCount) 
	else 
	  countString2 = ""
    end

    if fontStringTable and next(fontStringTable) then 
	  setFontStringTextsNil()
	  
	  fontStringTable.timerFontString:SetText(timerString)
      fontStringTable.affixFontString:SetText(affixString)
      fontStringTable.increaseTimersFontString:SetText(bonusTimerString)
      fontStringTable.deathsFontString:SetText(deathString)
      fontStringTable.bossesFontString:SetText(fullBossString)
      fontStringTable.countFontString:SetText(countString1..countString2..countString3)
	  fontStringTable.checkpointsFontString:SetText(checkpointsString)
    end
  end
  
end

function isTimeDataMissing(elapsedTime,missingCountInfo)

  local returnValue = false
  if missingCountInfo and missingCountInfo.missingTime then 
    local missingTime = missingCountInfo.missingTime
    for i, info in ipairs(missingTime) do
      if elapsedTime > missingTime[i][1] and elapsedTime <= missingTime[i][2] then  
        returnValue = true
        break
      end
    end
  end
  return returnValue

end


function isCountDataMissing(count,missingCountInfo)

  local returnValue = false
  if missingCountInfo and missingCountInfo.missingCount then 
    local missingCount = missingCountInfo.missingCount
    for i, info in ipairs(missingCount) do
      if count > missingCount[i][1] and count <= missingCount[i][2] then  
        returnValue = true
        break
      end
    end
  end
  return returnValue

end

function findCount(elapsedTime,countInfo,missingCountInfo) -- count when elapsed time

  local missing = false
  local count = nil

  if isTimeDataMissing(elapsedTime,missingCountInfo) then 
    missing = true
  end
  for i, info in ipairs(countInfo) do
    if countInfo[i][2] > elapsedTime then  -- if multiple entries with the same time finds the last entry
      break
    end
    count = countInfo[i][1]
  end
  
  return count,missing
  
end

function findTime(count,countInfo,missingCountInfo) --elapsed time when passed in count was first achieved

  local missing = false
  local elapsedTime = nil
  if isCountDataMissing(count,missingCountInfo) then 
    missing = true
  end
  for i, info in ipairs(countInfo) do
    if countInfo[i][1] >= count then
      elapsedTime = countInfo[i][2]
      break
    end
  end
   
  return elapsedTime,missing

end

function resolveFrames() -- called when enterring new zone and by challenge start event

  if isKeyActive() then 
       
    KeyRunnerFrame:Show()
	keyShown = true
	demoShown = false
	  
	if getConfig("hideBlizzardObjective") then 
	   
	  ObjectiveTrackerFrame:SetParent(EmptyHideFrame)
		
    else 
	   
	  ObjectiveTrackerFrame:SetParent(UIParent)
		
	end
	  
	   
  else 
	 
    KeyRunnerFrame:Hide()
    ObjectiveTrackerFrame:SetParent(UIParent)
		 
    keyShown = false
	demoShown = false
  end

end



function refreshFrame()

  if keyShown then 
    setFontStringsTexts(currentKey,compareKey)
  elseif demoShown then
    createDemoFrame()
  end
  setFontStringsAppearance()

end

function demoToggle()

  if not keyShown then
    if demoShown then 
      KeyRunnerFrame:Hide()
      demoShown = false
    else 

      createDemoFrame()
        
    end
  end
  
end



function createDemoFrame()

  createDemoKey()

  if getConfig("compareKeyLoad").disabled then 
    demoCompareKey = nil 
  else
    createDemoCompareKey()
  end

  setFontStringsTexts(demoKey,demoCompareKey)
  KeyRunnerFrame:Show()
  demoShown =  true

end


function getBossesKilledCount(bossInfo)

  local bossesKilled = 0
  for i, value in ipairs(bossInfo) do
    if value.completed then 
	  bossesKilled = bossesKilled +1
	end
  end

  return bossesKilled 

end

function getActiveKeyData()

  local keyID = 0
  local elapsedTime = 0
  local currentCount = 0
  local bossesKilled = 0

  if isKeyActive() then
    keyID = C_ChallengeMode.GetActiveChallengeMapID();
    _, elapsedTime =  GetWorldElapsedTime(1)
    local _, _, numSteps = C_Scenario.GetStepInfo()

    for i = 1, numSteps do
  
      local name, _, completed, curValue, finalValue, _, _, mobPoints= C_Scenario.GetCriteriaInfo(i); 
  
      if finalValue > 1 then  -- enemy forces
	 
	    if completed then 
	
	      currentCount = finalValue
	    else
	      local currentCountString = string.sub(mobPoints, 1, string.len(mobPoints) - 1)
          currentCount = tonumber(currentCountString)
	    end
	 
      else -- boss
	
	    if completed then
	      bossesKilled = bossesKilled + 1
        end
	    
      end
    end
  
  end

  return keyID,elapsedTime,currentCount,bossesKilled

end



function recoverSteps()

  if currentKey and isKeyActive() then
  
    currentKey["lastBossEndTime"] = 0
    currentKey["lastBossStartTime"] = 0
	local oldElapsedTime = currentKey.elapsedTime
	local _, elapsed_time = GetWorldElapsedTime(1)
	currentKey["elapsedTime"] = elapsed_time --  so updateSteps will create a datapoint with this elapsedTime rather than old one
	local _, _, numSteps = C_Scenario.GetStepInfo() 
	 
	if numSteps == 0 then -- after key finished then numSteps returns 0 also 0 before timer starts for some dungeons like shrine
	  return
	end
	
	if numSteps > currentKey.numSteps then -- count step added after key start or boss steps added after timer on shrine
	
	  currentKey.numSteps = numSteps  
	
	end
	 
	 
    for i = 1, currentKey.numSteps do
   
      local name, _, completed, curValue, finalValue, _, _, mobPoints= C_Scenario.GetCriteriaInfo(i); 
  
      if finalValue > 1 then  -- enemy forces
	    local oldCurrentCount = currentKey.currentCount
	    local newCount = 0
	    if completed then -- doesnt get set to 100%
	
	      newCount = finalValue
	    else
	      local currentCountString = string.sub(mobPoints, 1, string.len(mobPoints) - 1)
    
	      newCount = tonumber(currentCountString)
	    end
	   
		if newCount > oldCurrentCount then  -- count was updated when we were not there 
	      currentKey.dataMissing = true
		  table.insert(currentKey.missingCountInfo.missingCount, {oldCurrentCount,newCount}) 
		  table.insert(currentKey.missingCountInfo.missingTime, {oldElapsedTime,currentKey.elapsedTime})
        end	
		
	  else -- boss
	
	    if not currentKey.bossInfo[i] then
	      currentKey.bossInfo[i] = {}   
	      currentKey.bossInfo[i].name = name
	      currentKey.bossInfo[i].killTime = 0
	      currentKey.bossInfo[i].duration = 0
	      currentKey.bossInfo[i].completed = false
          currentKey.bossCount = currentKey.bossCount + 1
	    end 
	
	    if completed and not currentKey.bossInfo[i].completed then -- boss was killed while we were not there
		  currentKey.bossInfo[i].killTime = 0
	      currentKey.bossInfo[i].duration = 0
		  currentKey.dataMissing = true
		  currentKey.bossInfo[i].completed = completed
	
	    end
	    
      end
    end

end


end

function updateSteps()

  if currentKey and isKeyActive() then
  
    local _, _, numSteps = C_Scenario.GetStepInfo() 
	 
	if numSteps == 0 then -- after key finished then numSteps returns 0 also 0 before timer starts for some dungeons like shrine
	  return
	end
	
	if numSteps > currentKey.numSteps then -- count step added after key start or boss steps added after timer on shrine
	
	  currentKey.numSteps = numSteps  
	
    end
	 
	 
    for i = 1, currentKey.numSteps do
   
      local name, _, completed, curValue, finalValue, _, _, mobPoints= C_Scenario.GetCriteriaInfo(i); 
  
      if finalValue > 1 then  -- enemy forces
	    local oldCurrentCount = currentKey.currentCount
	    if completed then -- doesnt get set to 100%
	      currentKey["currentCount"] = finalValue 
	      currentKey ["finalCount"] = finalValue
	    else
	      local currentCountString = string.sub(mobPoints, 1, string.len(mobPoints) - 1)
          currentKey["currentCount"] = tonumber(currentCountString)   
	      currentKey ["finalCount"] = finalValue
	    end
	    if currentKey.currentCount > oldCurrentCount then  -- count was updated
		  currentKey.lastCountUpdate = currentKey.elapsedTime
	      table.insert(currentKey.countInfo, {currentKey.currentCount,currentKey.elapsedTime})
        end	
		
	  else -- boss
	
	    if not currentKey.bossInfo[i] then
	      currentKey.bossInfo[i] = {}   
	      currentKey.bossInfo[i].name = name
	      currentKey.bossInfo[i].killTime = 0
	      currentKey.bossInfo[i].duration = 0
	      currentKey.bossInfo[i].completed = false
          currentKey.bossCount = currentKey.bossCount + 1
	    end 
	
	    if completed and not currentKey.bossInfo[i].completed then -- new boss killed!!
	      currentKey.bossInfo[i].killTime = currentKey["elapsedTime"]
	
	      if currentKey["lastBossStartTime"]==0 or currentKey["lastBossEndTime"]==0 then 
		    currentKey.bossInfo[i].duration = 0
			currentKey.dataMissing = true
		  else 
			currentKey.bossInfo[i].duration = math.floor(currentKey["lastBossEndTime"] - currentKey["lastBossStartTime"])
			currentKey["lastBossEndTime"] = 0
			currentKey["lastBossStartTime"] = 0
		  end
	
	      currentKey.bossInfo[i].completed = completed
	    end
	    
      end
    end
  
    if fontStringTable then
      updateStepsTexts()
    end
  
  end

end



function updateTime(_,elapsedTime)

  -- skip until next second
  if not currentKey or (currentKey["elapsedTime"] == elapsed_time) then
    return
  end
 
  currentKey["elapsedTime"] = elapsedTime
	
  local deathcount, timeLost = C_ChallengeMode.GetDeathCount() 
  currentKey["deaths"] = deathcount 
	
  if fontStringTable then
    updateTimeTexts()
  end
	
end

function getCompareTooltipText(key)

  local returnString = ""
  
  if key then 
    returnString = getAffixString(key.level,key.affixIDs)
	returnString = returnString .. "\n" .. getTimerString(key.elapsedTime,key.timer)
	if not key.completed then 
	  returnString = returnString .. colorStringRed(" incomplete")
	end
    returnString = returnString .. "\n" .. getDeathString(key.deaths)
  end

  return returnString

end


function setCompareTooltip()

  if getConfig("showCompareTooltip")  then 
    if keyShown and compareKey then 
      GameTooltip:SetOwner(CompareTooltipFrame, "ANCHOR_TOP") 
      GameTooltip:SetText(getCompareTooltipText(compareKey),1,1,1,1) 

      GameTooltip:Show()
    elseif demoShown and demoCompareKey then 
	  GameTooltip:SetOwner(CompareTooltipFrame, "ANCHOR_TOP") 
      GameTooltip:SetText( getCompareTooltipText(demoCompareKey),1,1,1,1) 
	
      GameTooltip:Show()
	end
	
  end

end


function hideTooltip()
  GameTooltip:Hide()
end

function getFormattedPlayerName(player)

  local nameText = player.name
                 
  local colour = select(4, GetClassColor(player.class))

  nameText = WrapTextInColorCode(nameText, colour)
 
  return formatRoleIcon(nameText,player.role)
 
end

function getPlayersString(players)

  local playersString = ""

  for i, v in ipairs(players) do
    if i==1 then 
      playersString = getFormattedPlayerName(v) 
	else
	  playersString = playersString .. "\n" .. getFormattedPlayerName(v)
	end
	
  end

  return playersString

end





function isKeyActive () -- returns false if you are outside the key instance - returns true if key inserted but timer not started 
  
  local _, _, difficulty, _, _, _, _, _ = GetInstanceInfo()
  local _, elapsed_time = GetWorldElapsedTime(1)

  return C_ChallengeMode.IsChallengeModeActive() and difficulty == 8 and elapsed_time >= 0

end


function getTimerString(elapsedTime,timer)

  local timerString = ""
  local remainingTimeString = ""

  remainingTimeString = "  (" .. secondsToString(timer - elapsedTime) .. ")"
  timerString =  secondsToString(elapsedTime)  .. " / " .. secondsToString(timer) .. remainingTimeString

  return timerString
 
end



function getCompareAffixString(level)

  return " "..colorStringCompare("[+" .. level .. "]")

end


function getBonusTimerString(elapsedTime,timer)

  local timerString2 = ""
  local timerString3 = ""

  local remainingTimeString2 = ""
  local remainingTimeString3 = ""

  timerString2 = "+2=" .. secondsToString(math.floor(timer*0.8))
  remainingTimeString2 = " (" .. secondsToString(math.floor(timer*0.8 - elapsedTime)) .. ")"
  timerString2 = timerString2 .. remainingTimeString2

  timerString3 = "+3=" .. secondsToString(math.floor(timer*0.6))
  remainingTimeString3 = " (" .. secondsToString(math.floor(timer*0.6 - elapsedTime)) .. ")"
  timerString3 = timerString3 .. remainingTimeString3

  return timerString2.. " / " .. timerString3

end


function getAffixString(level,affixIDs)

  local affixString = ""

  for i, affixID in ipairs(affixIDs) do
    local affixName, affixDesc, affixNum = C_ChallengeMode.GetAffixInfo(affixID);
    affixString = affixString ..affixName .." "
  end

  return "[+" .. level .. "]" .." " .. affixString

end




function getDeathString(deaths)

  local deathString = deaths .. " deaths (-" .. secondsToString(deaths*5)..")" 

  return deathString

end






function getFullBossString(bossInfo,compareBossInfo,colorReplace,showDuration)
  
  local bossString = ""
  
  if bossInfo and next(bossInfo) then
  
    for i, value in ipairs(bossInfo) do
      local durationString = ""
      local killTimeString = ""
      local bossNameString = ""
      if value.completed then  
		if showDuration  then
          if value.duration>0 then 		  
		    durationString = colorStringTimer(secondsToString(value.duration),colorReplace) 
		    if compareBossInfo and next(compareBossInfo) and compareBossInfo[i] and compareBossInfo[i].completed then
			  if compareBossInfo[i].duration > 0 then 
				durationString = durationString .. secondsToStringColor(compareBossInfo[i].duration - value.duration)
			  elseif compareBossInfo[i].duration == 0 then
				durationString = durationString .. " " .. formatWarningIcon("")
			  end
			end
		  elseif value.duration ==0 then 
		    durationString = formatWarningIcon("")
		  end
		end
	    if value.killTime > 0 then 
	      killTimeString = colorStringTimer(secondsToString(value.killTime),colorReplace)
          if compareBossInfo and next(compareBossInfo) and compareBossInfo[i] and compareBossInfo[i].completed  then
            if compareBossInfo[i].killTime > 0 then 		 
              killTimeString = killTimeString .. secondsToStringColor(compareBossInfo[i].killTime - value.killTime)
            elseif compareBossInfo[i].killTime == 0 then
              killTimeString = killTimeString  .. " " .. formatWarningIcon("")			
		    end
          end
        elseif value.killTime == 0 then 
	      killTimeString = formatWarningIcon("")
	    end
	    
        bossNameString =    formatCompletedIcon(string.sub(value.name,1,-9))
      else
        bossNameString =    formatIncompleteIcon(string.sub(value.name,1,-9))
      end
	   
	  if string.len(durationString) > 0 then durationString = durationString .. " " end 
	  if string.len(killTimeString) > 0 then killTimeString = " "..killTimeString   end 
	   
      if i == 1 then
        bossString = durationString ..bossNameString ..killTimeString 
      else 
        bossString = bossString.."\n" .. durationString ..bossNameString..killTimeString 
      end
   
    end
  
  end
  
  return bossString
  
end


function getCountString(currentCount,finalCount)

  local percentString = ""
  local countDifference = ""

  if (currentCount >= finalCount) then
    percentString = "100%  " 
  else
    percentString = string.format("%.2f", currentCount/finalCount*100) .. "%  "
  end 

  if currentCount < finalCount then

    countDifference = " (" .. finalCount - currentCount .. ")"

  else 

    countDifference = " (0)"

  end

  if finalCount == 0 then
    percentString = "0.00%  "
  end
   
  if (currentCount >= finalCount) and finalCount ~= 0 then 
    return formatCompletedIcon("Enemy forces  ")..percentString .. currentCount  .. "/" .. finalCount ..countDifference
  else
    return formatIncompleteIcon("Enemy forces  ")..percentString .. currentCount  .. "/" .. finalCount ..countDifference
  end

end

function getCompareFindCountString(elapsedTime,compareCountInfo,compareMissingCountInfo,compareFinalCount)

  local returnString = ""
  local compareFindCount = nil
  local missing = false

  if compareCountInfo and next(compareCountInfo) then 
    compareFindCount,missing = findCount(elapsedTime,compareCountInfo,compareMissingCountInfo)
  end
	
  if missing then 
    returnString = " " ..formatWarningIcon("")..colorStringCompare("/" .. compareFinalCount)
  elseif compareFindCount then 
    returnString = " " .. colorStringCompare(compareFindCount .. "/" .. compareFinalCount)
  end
	
  return returnString
  
end


function getCompareFindTimeString (currentCount,missingCountInfo,compareCountInfo,compareMissingCountInfo,lastCountUpdate)

  local returnString = ""
  local compareFindTime = nil
  local missing = false 

  if isCountDataMissing(currentCount,missingCountInfo) then 
    missing = true
  elseif compareCountInfo and next(compareCountInfo) then 
    compareFindTime,missing = findTime(currentCount,compareCountInfo,compareMissingCountInfo)
  end
		
  if missing then -- no reliable time to compare
    returnString = " " ..formatWarningIcon("")
  elseif compareFindTime then 
    returnString  = secondsToStringColor(compareFindTime - lastCountUpdate)
  end
  
  return returnString    

end

function getCheckpointString(key1,key2,value,perc,colorReplace)  -- key1 = key and key2 = comparison key

  local returnString = ""
  local value1 = 0
  local value2 = 0
  
  if not perc then -- absolute
    value = math.floor(value)
	if value > key1.finalCount then 
	  value1 = key1.finalCount
	else
	  value1 = value
	end
    
	local t1,missing1 = findTime(value1,key1.countInfo,key1.missingCountInfo)
    if t1 then 
	  returnString =  formatCompletedIcon(value1) .. "/" .. value1
	  if missing1 then 
		returnString =  returnString.. " "..formatWarningIcon("")
	  else
		returnString =  returnString .. " "..colorStringTimer(secondsToString (t1),colorReplace) 
	  end
	else 
	  returnString =  formatIncompleteIcon(key1.currentCount) .. "/" .. value1
	end
		
    if key2 and key2.finalCount and key2.countInfo and t1 and not missing1 then -- t1 has to be there and reliable(t1 will be nil for incomplete)
	  if value > key2.finalCount then 
	    value2 = key2.finalCount
	  else
	    value2 = value
	  end
	  local t2,missing2 = findTime(value2,key2.countInfo,key2.missingCountInfo)
	  if missing2 then 
		returnString = returnString .. " "..formatWarningIcon("") 
	  elseif t2 then 
		returnString = returnString .. secondsToStringColor(t2-t1) 
	  end
	end
  
  else  -- percent
    value =  tonumber(string.format("%.2f", value))
	if value > 100 then value = 100 end
    local countValue1 = math.floor(value/100*(key1.finalCount))
    local t1,missing1 = findTime(countValue1,key1.countInfo,key1.missingCountInfo)
      
	if t1 then 
	  returnString =  formatCompletedIcon(value) .. "%/" .. value .. "%"
	  if missing1 then 
		returnString =  returnString.. " "..formatWarningIcon("")
	  else
		returnString =  returnString .. " "..colorStringTimer(secondsToString (t1),colorReplace) 
	  end
	else 
	  local curPerc = string.format("%.2f", (key1.currentCount/key1.finalCount * 100))
	  returnString =  formatIncompleteIcon(curPerc) .. "%/".. value .. "%"
	end
	   
	if key2 and key2.finalCount and key2.countInfo and t1 and not missing1 then -- t1 has to be there and reliable(t1 will be nil for incomplete)
	  local countValue2 = math.floor(value/100*(key2.finalCount))
	  local t2,missing2 = findTime(countValue2,key2.countInfo,key2.missingCountInfo)
	  if missing2 then
		returnString = returnString .. " "..formatWarningIcon("")
	  elseif t2 then 
		returnString = returnString .. secondsToStringColor(t2-t1) 
      end
	end
  end
  
  return returnString  
end

function getFullCheckpointString(key1,key2,colorReplace)

  local returnString = ""
  local checkpoint1 = getConfig("checkpoint1")
  local checkpoint2 = getConfig("checkpoint2")
  local checkpoint1String = ""
  local checkpoint2String = ""
  
  
  if not checkpoint1.disabled and checkpoint1.value ~= nil and key1.finalCount > 0 then
    if checkpoint1.type == 1 then -- absolute
	  checkpoint1String = getCheckpointString(key1,key2,checkpoint1.value,false,colorReplace)
    else  -- percent
	  checkpoint1String = getCheckpointString(key1,key2,checkpoint1.value,true,colorReplace)
	end
  end
  
  if not checkpoint2.disabled and checkpoint2.value ~= nil and key1.finalCount > 0 then
    if checkpoint2.type == 1 then -- absolute
	  checkpoint2String = getCheckpointString(key1,key2,checkpoint2.value,false,colorReplace)
    else  -- percent
	  checkpoint2String = getCheckpointString(key1,key2,checkpoint2.value,true,colorReplace)
	end
  end

  if string.len(checkpoint1String) > 0 and  string.len(checkpoint2String) > 0 then 
    returnString = checkpoint1String .. "\n" .. checkpoint2String 
  elseif string.len(checkpoint2String) > 0 then 
    returnString = checkpoint2String
  elseif string.len(checkpoint1String) > 0 then
    returnString = checkpoint1String
  end

  return returnString
  
  
  
end


function secondsToStringColor(timeCompare)

  local timeCompareString = ""
  if timeCompare < 0 then 
    timeCompareString = colorStringRed(" -" .. secondsToString(math.abs(timeCompare)))  -- color red
  elseif timeCompare == 0 then 
    timeCompareString = colorStringGreen(" " .. secondsToString(math.abs(timeCompare)))   -- color green
  elseif timeCompare >0 then 
    timeCompareString = colorStringGreen(" +" .. secondsToString(math.abs(timeCompare)))   -- color green
  end
  return timeCompareString

end

function secondsToString(seconds)

 

  local negativeString = ""
  local returnString = ""
  
  if seconds ~=nil then
    if ( seconds < 0 ) then
      seconds = math.abs(seconds)
      negativeString = "-"
    end 

    local minutes = math.floor(seconds/60);
    local sec = seconds - (minutes * 60);
  
    if (sec < 10) then
      sec = string.format("0%d", sec);
    else
      sec = string.format("%d", sec);
    end
  
    returnString = negativeString..minutes .. ":" .. sec
  
  end
  
  return returnString
  
end

function formatWarningIcon(str)
  return string.format("|T%s:%s|t%s", "Interface\\AddOns\\KeyRunner\\Media\\warning", "0:0:0:0", str)
end


function formatCompletedIcon(str)
  return string.format("|T%s:%s|t%s", "Interface\\AddOns\\KeyRunner\\Media\\completed", "0:0:0:-1", str)
end

function formatIncompleteIcon(str)
  return string.format("|T%s:%s|t%s", "Interface\\AddOns\\KeyRunner\\Media\\incomplete", "0:0:1:-1", str)
end

function formatRoleIcon(str,role)

  local roleIcons = {
	["TANK"] = "Interface\\AddOns\\KeyRunner\\Media\\tank",
	["HEALER"] = "Interface\\AddOns\\KeyRunner\\Media\\healer",
	["DAMAGER"] = "Interface\\AddOns\\KeyRunner\\Media\\dps"
  }

  return string.format("|T%s:%s|t%s", roleIcons[role], "0:0:0:0", str)
  
end


function getOptionsTimerColor()

  local color = {["r"] = 0.8,["g"] = 0.8,["b"] = 0}
  return  color

end

function getBossTimerColor()
  
  local fonts = getConfig("fonts")
  local color = fonts.fonts9
  return color

end

function getCheckpointTimerColor()
  
  local fonts = getConfig("fonts")
  local color = fonts.fonts10
  return color

end

function  colorStringTimer(str,color)

  local stringRep = ""

  stringRep = "|cFF" .. RGBPercToHex(color.r,color.g,color.b) ..str.. "|r"

  return stringRep

end

function colorStringCompare(str)

  local stringRep = ""

  local fonts = getConfig("fonts")
  local color = fonts.fonts8

  stringRep = "|cFF" .. RGBPercToHex(color.r,color.g,color.b) ..str.. "|r"

  return stringRep


end



function colorStringRed(str)
  return "|cFFFF0000" ..str .. "|r"  -- alpha channel ignored must be set to FF
end

function colorStringGreen(str)
  return "|cFF00FF00" ..str .. "|r"  -- alpha channel ignored must be set to FF
end

function RGBPercToHex(r, g, b)
  r = r <= 1 and r >= 0 and r or 0
  g = g <= 1 and g >= 0 and g or 0
  b = b <= 1 and b >= 0 and b or 0
  return string.format("%02x%02x%02x", r*255, g*255, b*255)
end